import os
import time
import json
import shutil
import zipfile
import pathlib
import requests
from .errors import Errors

#
# Copyright (c) 2021 Pharmacelera S.L.
# All rights reserved.
#
# Description: Pharmacelera API calls
#


class Calls:
    """Pharmacelera API calling class."""

    def __init__(self, params):
        """Initialize configured parameters"""
        self.__errors = Errors() or None
        self.__path = params.get("MODEL_NAME") or ""
        self.__port = os.getenv("PORT") or params.get("PORT")
        self.__url = f"{params.get('URI')}:{self.__port}"

    def upload(self, files):
        """Upload binary and license inside container."""
        url = f"{self.__url}/uploadBinary"
        response = requests.request(
            "POST",
            url,
            files=files,
        )
        result = json.loads(response.text)
        return result

    def launch(self, endpoint, payload, files):
        """Launch an experiment with given data."""
        url = f"{self.__url}/{endpoint}"
        try:
            response = requests.request(
                "POST",
                url,
                data=payload,
                files=files,
            )
            result = json.loads(response.text)
            return result
        except Exception as exp:
            self.__errors.requestException(exp)

    def get(self, id):
        try:
            """Get experiment status data."""
            url = f"{self.__url}?id={id}"
            response = requests.request(
                "GET",
                url,
            )
            result = json.loads(response.text)
            return result
        except Exception as exp:
            self.__errors.requestException(exp)

    def list(self, query_param):
        try:
            """List all experiments."""
            url = f"{self.__url}/list?{query_param}"
            response = requests.request(
                "GET",
                url,
            )
            result = json.loads(response.text)
            return result
        except Exception as exp:
            self.__errors.requestException(exp)

    def download(self, id, folder, path=None, outputs="outputs", count=0):
        """Download a completed experiment."""
        if count == 0:
            print("Downloading the latest experiment files...")
        url = f"{self.__url}/download?id={id}"
        zipName = folder + ".zip"
        try:
            with requests.request("GET", url, stream=True) as r:
                r.raise_for_status()
                if r.headers.get("content-type").startswith("application/json"):
                    time.sleep(count)
                    if count == 20:
                        print(json.loads(r.text).get("body").get("msg"))
                    count += 1
                    if count <= 20:
                        self.download(id, folder, path, outputs, count=count)
                    return
                zip_path = os.path.join(outputs, self.__path)
                pathlib.Path(os.path.join(path, zip_path)).mkdir(
                    parents=True, exist_ok=True
                )
                with open(os.path.join(zip_path, zipName), "wb") as f:
                    for chunk in r.iter_content(chunk_size=8192):
                        f.write(chunk)
                if r.headers.get("content-type") == "application/zip":
                    self.extractFiles(zip_path, zipName)
                    print(f"{zipName} Downloaded")
        except Exception as exp:
            self.__errors.downloadFilesException(exp)

    def kill(self, id):
        """Stop the running experiment manually."""
        try:
            url = f"{self.__url}/kill?id={id}"
            response = requests.request("GET", url)
            result = json.loads(response.text)
            return result
        except Exception as exp:
            self.__errors.requestException(exp)

    def remove(self, id):
        """Remove experiment, stop first if running and delete data and files."""
        try:
            url = f"{self.__url}/remove?id={id}"
            response = requests.request("DELETE", url)
            result = json.loads(response.text)
            return result
        except Exception as exp:
            self.__errors.requestException(exp)

    def extractFiles(self, zip_path, zipName):
        """Extract sdf files from compressed files."""
        dirname = pathlib.Path().absolute()
        with zipfile.ZipFile(os.path.join(dirname, zip_path, zipName)) as zipped:
            for file in zipped.namelist():
                if (
                    file.endswith((".sdf", "mol2"))
                    and file.find("erroneous") < 0
                    and file.find("reference_molecules") < 0
                ):
                    filename = zipName.rsplit(".", 1)[0]
                    path = os.path.join(dirname, "tmp", filename)
                    zipped.extract(file, path)
                    ext = "." + file.split(".")[-1]
                    target = filename + ext
                    shutil.move(
                        os.path.join(path, file),
                        os.path.join(dirname, "tmp", target),
                    )
                    os.rmdir(path)

    def clearFiles(self):
        """Clear input and output files from container
        after finish and download the experiment.
        """
        url = f"{self.__url}/clearAll"
        response = requests.request("DELETE", url)
        result = json.loads(response.text)
        return result



#!/usr/bin/env python

import sys
import argparse
from typing import List
from pathlib import Path
from .utils import Utils
from .utils import Colors
from .services import Calls


class ArgumentParser:
    def __init__(self) -> None:
        self.init_arguments()

    def init_arguments(self):
        """Define cmd input arguments."""
        parser = argparse.ArgumentParser("")
        parser.add_argument(
            "--option",
            help="Type of the command to call Pharmacelera endpoint. Possible options: get, list, download, kill, remove, upload. Default: get. Note: list option supports filter by status, eg: list?status=running.",
            type=str,
            default="get",
        )
        parser.add_argument("--id", help="Unique id of the experiment", type=str)
        parser.add_argument(
            "--uri", help="Host name of the endpoint. Default: http://localhost", type=str, default="http://localhost"
        )
        parser.add_argument("--port", help="endpoint port number. Default: 8080", type=str, default="8080")
        parser.add_argument(
            "--type",
            help="Type of the binary to upload: [pharmqsar, pharmscreen]. default: pharmqsar",
            default="pharmqsar",
            type=str,
        )
        parser.add_argument("--binary", help="Path and filename of the binary file to upload", type=str)
        parser.add_argument(
            "--license",
            help="Path and filename of the license file to upload.",
            type=str,
        )
        self.args = parser.parse_args()


class AppConfiguration:
    client: Calls
    utils: Utils
    args: argparse.Namespace
    path: str = Path().absolute()
    printables: List = ["id", "msg", "status", "progress", "end_date", "experiment_type", "folder"]

    def __init__(self, args) -> None:
        params = {"PORT": args.port, "URI": args.uri}
        self.client = Calls(params)
        self.utils = Utils()
        self.args = args

    def get(self):
        id = self.args.id
        response = self.client.get(id)
        if not response:
            return
        if response.get("statusCode", None) == 200:
            body = response.get("body")
            self.print([body])
            return body
        else:
            print(f"{Colors.FAIL}No experiment found with the given id.{Colors.ENDC}")
        return None

    def list(self, query_param):
        print(f"{Colors.BOLD}Listing experiments ...{Colors.ENDC}")
        response = self.client.list(query_param)
        if not response:
            return
        body = response.get("body", None)
        if len(body):
            self.print(body)
        else:
            print(f"{Colors.FAIL}No experiment found.{Colors.ENDC}")

    def download(self):
        print(f"{Colors.BOLD}Starts downloading ...{Colors.ENDC}")
        experiment = self.get()
        if not experiment:
            return
        if experiment.get("status") not in ("finished", "stopped", "error"):
            print(f"{Colors.FAIL}Experiment is still running{Colors.ENDC}")
            return
        folder = experiment.get("folder").split("/")[-1]
        id = self.args.id
        self.client.download(id, folder, self.path, outputs="")
        print(f"{Colors.BOLD}Download finished.{Colors.ENDC}")

    def kill(self):
        print(f"{Colors.BOLD}Killing experiment ...{Colors.ENDC}")
        id = self.args.id
        experiment = self.get()
        if not experiment:
            return
        response = self.client.kill(id)
        print(response.get("body", None))
        print(f"Done.")

    def remove(self):
        print(f"{Colors.BOLD}Removing experiment ...{Colors.ENDC}")
        id = self.args.id
        experiment = self.get()
        if not experiment:
            return
        self.client.remove(id)
        print(f"Done.")

    def upload(self):
        print(f"{Colors.BOLD}Uploading binary and/or license ...{Colors.ENDC}")
        binary_type = self.args.type
        binary = self.args.binary
        license = self.args.license
        if not binary and not license:
            print(f"{Colors.FAIL}Please define binary and/or license path and filename.{Colors.ENDC}")
            return
        files = self.utils.define_files(binary_type, binary, license)
        if not files:
            return
        response = self.client.upload(files)
        body = response.get("body")
        if body:
            print(f'{Colors.OKGREEN}{body.get("msg")}{Colors.ENDC}')
        else:
            print(response)

    def print(self, experiment_lists):
        for index, experiment in enumerate(experiment_lists):
            print(f"{Colors.OKBLUE}Count: {index + 1}{Colors.ENDC}")
            print(120 * "-")
            single_list_experiment = list(experiment.items())
            single_list_experiment.sort(key=self.get_length)
            sorted_single_experiment = {str(elem[0]): str(elem[1]) for elem in single_list_experiment}
            colon_distance = str(len(list(sorted_single_experiment)[-1]))
            row_format = "{:<" + colon_distance + "} : {:<30}"
            for key, value in sorted_single_experiment.items():
                if key in self.printables:
                    print(f"{Colors.OKGREEN}{row_format.format(key, value)}{Colors.ENDC}")
            print(120 * "-" + "\n")

    def get_length(self, key):
        return len(key[0])

    def redirect(self, option: str):
        option = option.split("?")
        query_param = option[1] if len(option) > 1 else None
        if option[0] == "get":
            return self.get()
        elif option[0] == "list":
            return self.list(query_param)
        elif option[0] == "download":
            return self.download()
        elif option[0] == "kill":
            return self.kill()
        elif option[0] == "remove":
            return self.remove()
        elif option[0] == "upload":
            return self.upload()
        else:
            print("Option value is unknown.")


def run():
    parser = ArgumentParser()
    app = AppConfiguration(parser.args)
    if len(sys.argv) > 1:
        app.redirect(app.args.option)
    else:
        print("No argument parameter is present. Please, use --help to get more information")


#!/usr/bin/env python

import os
import sys
import time
import yaml
import time
import uuid
import boto3
import shutil
import logging
import zipfile
from pathlib import Path
from .errors import Errors
from .services import Calls
from .utils import Colors, Utils
from urllib.parse import urlparse

#
# Copyright (c) 2021 Pharmacelera S.L.
# All rights reserved.
#
# Description: Script to launch multiple PharmQSAR experiments and select the best model
#
# Usage: Define the desired experiments in the config.yaml file and run this script
#


class Configs:
    """Read pharmQSAR config file, and parse Yaml configs"""

    path: str = Path().absolute()

    def __init__(self) -> None:
        self.error = Errors()

    def get_from_s3(self, statics):
        """Download selected model from S3.

        Args:
            config (dict): Experiment configuration from yaml
        """
        filename = statics.get("FILENAME", None)
        if not filename:
            self.error.fileNameException()
        s3_bucket = statics.get("S3_BUCKET", None)
        if not s3_bucket:
            self.error.bucketNameException()
        s3_folder = statics.get("S3_FOLDER", None)
        if not s3_folder:
            self.error.S3FolderPathException()
        zip_path = os.path.join(self.path, "inputs")
        if not os.path.isdir(zip_path):
            Path(zip_path).mkdir(parents=True, exist_ok=True)
        try:
            self.download(
                s3_bucket, s3_folder + "/" + filename, os.path.join(zip_path, "tmp.zip")
            )
            folder, _, filenames = next(os.walk(zip_path))
            for f in filenames:
                if f.endswith(".zip"):
                    with zipfile.ZipFile(os.path.join(folder, f)) as zipped:
                        zipped.extractall(zip_path)
            shutil.move(
                os.path.join(zip_path, "custom.yaml"),
                os.path.join(self.path, "custom.yaml"),
            )
            os.remove(os.path.join(zip_path, "tmp.zip"))
            logging.info("Downloaded files and config.")
            return True
        except Exception as exp:
            self.error.S3Exception(exp)
            return False

    def download(self, bucket, path, local_path):
        """Downloads file from S3.
        Args:
            bucket(str): S3 bucket name
            path(str): filename with path to file
            local_path(str): path and filename in local to store the file
        """
        s3 = boto3.resource("s3")
        try:
            s3.Bucket(bucket).download_file(path, local_path)
            logging.info(f"{path} downloaded.")
        except Exception as exp:
            self.error.S3Exception(exp)

    def read(self, configFile):
        """Read and parse file."""
        try:
            with open(configFile, "r") as s:
                return yaml.safe_load(s)
        except yaml.parser.ParserError as exp:
            self.error.yamlParseException(exp)
        except Exception as exp:
            logging.error(exp)
        return None

    def custom_config(self):
        """Read and parse file."""
        try:
            with open("custom.yaml", "r") as s:
                return yaml.safe_load(s)
        except Exception as exp:
            logging.error(exp)
        return None

    def s3_config(self, statics):
        """Replace input molecule name of experiment1.
        Args:
            dict: default config.yaml configuration of Case4
        """
        config = self.custom_config()
        input_name = statics.get("EXP_1_INPUT", None)
        if not input_name:
            self.error.inputFileNameException()
        config["experiment1"]["files"]["input"] = input_name
        return config

    def get_yaml_config(self):
        """If yaml config path is S3, then download and parse the filename,
        config is in local, just return the filename.
        """
        args = sys.argv
        if len(args) < 2:
            print("Please define yaml configuration file")
            sys.exit(0)
        return self.s3_path_finder(args[1])

    def s3_path_finder(self, load_file: str):
        if load_file.startswith("s3://"):
            bucket_name, filepath, conf_name = self.parse_s3_url(load_file)
            logging.info(
                f"{Colors.BOLD}{Colors.OKBLUE}{conf_name} file will be downloading from S3{Colors.ENDC}\n"
            )
            self.download(bucket_name, filepath, conf_name)
            return conf_name
        else:
            logging.info(
                f"{Colors.BOLD}{Colors.OKBLUE}{load_file} file will be loading from local.{Colors.ENDC}\n"
            )
            return load_file

    def parse_s3_url(self, url):
        """Parse S3 url and extract bucket name, path and file name.
        Args:
            url(str): S3 url path
        Returns:
            (tupple): bucket name, file path, file name
        """
        parser = urlparse(url, allow_fragments=False)
        bucket_name = parser.netloc
        filepath = parser.path[1:]
        conf_name = os.path.basename(filepath)
        return (bucket_name, filepath, conf_name)


class Initialize:
    """Define params."""

    utils: Utils = None
    client: Calls = None
    config: Configs = None
    path: str = Path().absolute()

    def __init__(self, config):
        self.utils = Utils()
        self.error = Errors()
        self.config = Configs()
        self._set_params(config)
        self.client = Calls(config)
        self.retry = 5
        self.timeout = 1

    def _set_params(self, config):
        # define which binary will be executing
        self.__endpoint = config.pop("endpoint")

        # set all input files
        files = config.pop("files")
        for key, file in files.items():
            filename = self.config.s3_path_finder(file)
            files[key] = filename
        try:
            self.__files = {
                key: open(os.path.join(self.path, value), "rb")
                for (key, value) in files.items()
            }
        except Exception:
            self.error.inputFilesException(files)
        self.__parameters = config

    def launch(self):
        """Call launch API endpoint with defined parameters
        to start the experiment.
        """
        response = self.client.launch(self.__endpoint, self.__parameters, self.__files)
        if response:
            if response["statusCode"] == 200:
                exp = response["body"]
                return self.exp_getter(exp["id"])
            else:
                logging.error(
                    f"{Colors.FAIL}Experiment failed, {response.get('body')}{Colors.ENDC}"
                )
        else:
            if self.retry > 0:
                logging.error(
                    f"Experiment could not start, retrying again in {self.timeout} seconds"
                )
                time.sleep(self.timeout)
                self.retry -= 1
                self.timeout = self.timeout * 2
                self.launch()
        for _, open_file in self.__files.items():
            open_file.close()
        return False

    def exp_getter(self, id):
        """Watch status and progress of the experiment and
        download files when it is done.
        """
        logging.info(f"experiment id: {id}")
        status = "pending"
        while True:
            if status != "finished" and status != "error":
                response = self.client.get(id)
                id = response["body"]["id"]
                env = response["body"].get("env", "unknown")
                status = response["body"]["status"]["status"]
                command = response["body"].get("args", None)
                folder = self.utils.get_name(command)
                if status == "running":
                    progress = response["body"]["status"]["progress"]
                    logging.info(
                        f"{Colors.OKBLUE}Status: {status}, Progress: {progress}{Colors.ENDC}"
                    )
                    time.sleep(10)
                elif status == "pending":
                    logging.info(
                        f"Status: {status}, experiment id: {id}, running environment: {env}"
                    )
                    time.sleep(10)
                elif status == "finished":
                    logging.info(
                        f"{Colors.OKGREEN}{Colors.UNDERLINE}Experiment finished!{Colors.ENDC}"
                    )
                elif status == "stopped":
                    logging.warn(
                        f"{Colors.WARNING}Experiment stopped manually.{Colors.ENDC}"
                    )
                    self.client.download(id, folder, self.path)
                    return False
                elif status == "error":
                    logging.error(
                        f"{Colors.FAIL}Experiment finished with Error.{Colors.ENDC}"
                    )
                    logging.error(
                        f"{Colors.FAIL}{response['body']['status']}{Colors.ENDC}"
                    )
                    self.client.download(id, folder, self.path)
                    return False
                time.sleep(1)
            else:
                self.client.download(id, folder, self.path)
                logging.info(
                    f"Experiment with id: {id} downloaded in outputs."
                )
                return True

    def clear(self):
        try:
            shutil.rmtree(os.path.join(self.path, "tmp"))
            os.remove(os.path.join(self.path, "custom.yaml"))
        except Exception:
            pass


def run():
    logging.basicConfig(level=logging.INFO)
    app_config = Configs()
    config_file_name = app_config.get_yaml_config()
    config = app_config.read(config_file_name)
    start = time.time()
    if config:
        statics = config.pop("static")
        get_conf = statics.get("GET_CONF_FROM_S3", None)
        if get_conf:
            if not app_config.get_from_s3(statics):
                sys.exit()
            config = app_config.s3_config(statics)
        statics["MODEL_NAME"] = f'{uuid.uuid4().hex}_{statics.get("MODEL_NAME", "")}'
        if not config.keys():
            logging.error(
                f"{Colors.FAIL}No experiment parameter is defined.{Colors.ENDC}"
            )
            sys.exit(0)
        for k, v in config.items():
            # Run each experiment synchronously from configuration
            logging.info(f" - - - - Starting experiment with {k} params - - - -")
            statics["jobName"] = v["name"]
            params = {**statics, **v}
            init = Initialize(params)
            result = init.launch()
            if not result and not get_conf:
                sys.exit(0)

        if statics.get("UPLOAD_S3", None):
            # select best model
            init.utils.best_model_selector(init.path, statics, config)
        # clear input and out folders from API container
        init.client.clearFiles()
        # clear temporary files
        init.clear()
        logging.info("done")
    else:
        logging.error("Could not load experiments from config")
    print(f"total elapsed time: {time.time() - start:.5f}")



import os
import glob
import shutil
import unittest
import subprocess
from pathlib import Path


class TestLaunch(unittest.TestCase):
    """pharmacelera-launch test cases."""

    path = Path().absolute()

    def tearDown(self) -> None:
        self.remove("input_molecules.sdf", "activities.csv")
        shutil.rmtree(os.path.join(self.path, "inputs"), ignore_errors=True)
        shutil.rmtree(os.path.join(self.path, "outputs"), ignore_errors=True)

    def test_A_yaml_config_arg_not_present(self):
        """Executing pharmacelera-launch command without argument of config file should return error message for missing yaml config."""
        print(self.shortDescription())
        out = subprocess.check_output("pharmacelera-launch")
        self.assertEqual(str(out, "UTF-8").rstrip(), "Please define yaml configuration file")

    def test_B_yaml_config_file_not_exist(self):
        """Executing command with config file argument that does not exist should return not found error."""
        print(self.shortDescription())
        process = subprocess.run(
            ["pharmacelera-launch", "configNotExist.yaml"],
            capture_output=True,
            text=True,
        )
        print(process.stderr)
        err = process.stderr.split("\n")[0]
        err_2 = process.stderr.split("\n")[2]
        self.assertEqual(err, "INFO:root:\x1b[1m\x1b[94mconfigNotExist.yaml file will be loading from local.\x1b[0m")
        self.assertEqual(err_2, "ERROR:root:[Errno 2] No such file or directory: 'configNotExist.yaml'")

    def test_C_yaml_config_file_s3_not_exist(self):
        """Executing command with s3 path that file is not exist to download should return not found error."""
        print(self.shortDescription())
        process = subprocess.run(
            ["pharmacelera-launch", "s3://pharmacelera-launch/pha-testing/configNotExist.yaml"],
            capture_output=True,
            text=True,
        )
        print(process.stderr)
        out_with_color = process.stderr.split("\n")[0]
        err_1 = process.stderr.split("\n")[-2]
        err_2 = process.stderr.split("\n")[-3]
        self.assertEqual(out_with_color, "INFO:root:\x1b[1m\x1b[94mconfigNotExist.yaml file will be downloading from S3\x1b[0m")
        self.assertEqual(err_1, "ERROR:root:Could not load experiments from config")
        self.assertEqual(err_2, "ERROR:root:[Errno 2] No such file or directory: 'configNotExist.yaml'")

    def test_D_incorrectly_formatted_yaml_config_file(self):
        """Executing command with a config file that yaml is not formatted correctly."""
        print(self.shortDescription())
        process = subprocess.run(
            ["pharmacelera-launch", "testMocks/config_incorrect.yaml"], capture_output=True, text=True
        )
        print(process.stderr)
        err_with_color = process.stderr.split("\n")[2]
        self.assertEqual(err_with_color, "ERROR:root:\x1b[91mYaml config file is malformed.\x1b[0m")

    def test_E_no_experiment_params_in_yaml_config_file(self):
        """Executing command with a config file that has no any experiment configuration."""
        print(self.shortDescription())
        process = subprocess.run(
            ["pharmacelera-launch", "testMocks/config_no_experiment.yaml"], capture_output=True, text=True
        )
        print(process.stderr)
        err_with_color = process.stderr.split("\n")[2]
        self.assertEqual(err_with_color, "ERROR:root:\x1b[91mNo experiment parameter is defined.\x1b[0m")

    def test_F_parameter_calculation(self):
        """Run a pharmQSAR experiment with config1"""
        print(self.shortDescription())
        outputs = []
        for line in self.execute(["pharmacelera-launch", "testMocks/config1.yaml"]):
            if line.strip():
                outputs.append(line.strip())
        self.assertTrue(Path(self.path, "outputs").is_dir())
        # Make sure 2 output files are downloaded
        outputs = glob.glob(os.path.join(self.path, "outputs/*/*"))
        self.assertEqual(len(outputs), 2)
        for file in outputs:
            self.assertTrue(Path(self.path, file).is_file())
            self.assertGreater(file.index("test_case_"), 0)
            self.assertGreater(Path(self.path, file).stat().st_size, 0)
        # Make sure best model zip file is generated
        generated_output = False
        for file in glob.glob(str(self.path) + "/*"):
            if file.endswith("_TEST_MODEL.zip"):
                generated_output = True
                output_file = file
        # delete zip file after test
        self.assertTrue(generated_output)
        if generated_output:
            self.remove(output_file)

    def test_J_model_generation_from(self):
        """Download experiment file from S3 and run model generate."""
        print(self.shortDescription())
        outputs = []
        for line in self.execute(["pharmacelera-launch", "testMocks/config4.yaml"]):
            if line.strip():
                outputs.append(line.strip())
        inputs = os.path.join(self.path, "inputs")
        self.assertTrue(os.path.isdir(inputs))

        # Make sure 2 output files are downloaded
        outputs = glob.glob(os.path.join(self.path, "outputs/*/*"))
        self.assertEqual(len(outputs), 2)
        for file in outputs:
            self.assertTrue(Path(self.path, file).is_file())
            self.assertGreater(file.index("test_case_"), 0)
            self.assertGreater(Path(self.path, file).stat().st_size, 0)

    def execute(self, cmd):
        """Executing command line with subprocess."""
        popen = subprocess.Popen(cmd, stdout=subprocess.PIPE, universal_newlines=True)
        for stdout_line in iter(popen.stdout.readline, ""):
            yield stdout_line
        popen.stdout.close()
        return_code = popen.wait()
        if return_code:
            raise subprocess.CalledProcessError(return_code, cmd)

    def remove(self, *args):
        """Remove files and ignore error.
        args: iterable strings
        """
        for file in args:
            try:
                os.remove(os.path.join(self.path, file))
            except OSError:
                pass



import os
import glob
import yaml
import shutil
import unittest
import subprocess
from pathlib import Path


class TestCMDCommands(unittest.TestCase):
    """pharmacelera-launch test cases."""

    statics: dict
    path = Path().absolute()

    @classmethod
    def setUpClass(cls) -> None:
        try:
            with open(os.path.join(cls.path, "testMocks", "config.yaml"), "r") as filename:
                config = yaml.safe_load(filename)
                cls.statics = config.get("static")
        except yaml.parser.ParserError:
            print("Could not parse yaml config")
        except Exception as exp:
            print(exp)

    def tearDown(self) -> None:
        self.remove("input_molecules.sdf", "activities.csv")
        shutil.rmtree(os.path.join(self.path, "inputs"), ignore_errors=True)
        shutil.rmtree(os.path.join(self.path, "outputs"), ignore_errors=True)

    def test_A_run_cmd_command_no_arg(self):
        """Executing CMD command without any arguments."""
        print(self.shortDescription())
        out = subprocess.check_output("pharmacelera-cmd")
        self.assertEqual(
            str(out, "UTF-8").rstrip(), "No argument parameter is present. Please, use --help to get more information"
        )

    def test_B_run_cmd_no_id_value(self):
        """Executing CMD command with empty id."""
        print(self.shortDescription())
        process = subprocess.run(
            ["pharmacelera-cmd", "--id"],
            capture_output=True,
            text=True,
        )
        err = process.stderr.split("\n")[-2].strip()
        self.assertEqual(err, ": error: argument --id: expected one argument")

    def test_C_run_cmd_no_host_connection(self):
        """Executing CMD with no connection to host."""
        print(self.shortDescription())
        process = subprocess.run(
            ["pharmacelera-cmd", "--id", "xxx", "--port", "5000"],
            capture_output=True,
            text=True,
        )
        err = process.stderr.split("\n")[0]
        err_desc = process.stderr.split("\n")[1]
        self.assertEqual(err, "ERROR:root:Request failed")
        self.assertGreater(err_desc.index("Failed to establish a new connection"), 0)

    def test_D_run_cmd_list_with_wrong_status(self):
        """Executing CMD list experiments with wrong status."""
        print(self.shortDescription())
        process = subprocess.run(
            [
                "pharmacelera-cmd",
                "--option",
                "list?status=unknown",
                "--port",
                self.statics["PORT"],
                "--uri",
                self.statics["URI"],
            ],
            capture_output=True,
            text=True,
        )
        out_with_color = process.stdout.split("\n")[1]
        self.assertEqual(out_with_color, "\x1b[91mNo experiment found.\x1b[0m")

    def test_E_run_cmd_list_with_finished_status(self):
        """Executing CMD list all finished experiments."""
        print(self.shortDescription())
        process = subprocess.run(
            [
                "pharmacelera-cmd",
                "--option",
                "list?status=finished",
                "--port",
                self.statics["PORT"],
                "--uri",
                self.statics["URI"],
            ],
            capture_output=True,
            text=True,
        )
        self.assertGreater(process.stdout.index("finished"), 0)

    def test_G_run_example_experiment(self):
        """Launching experiment for further cmd command testings."""
        print(self.shortDescription())
        outputs = []
        for line in self.execute(["pharmacelera-launch", "testMocks/config.yaml"]):
            if line.strip():
                outputs.append(line.strip())
        self.__class__.exp_id = outputs[0]

    def test_H_get_single_experiment(self):
        """Executing CMD get single experiment with id."""
        print(self.shortDescription())
        process = subprocess.run(
            [
                "pharmacelera-cmd",
                "--id",
                self.__class__.exp_id,
                "--port",
                self.statics["PORT"],
                "--uri",
                self.statics["URI"],
            ],
            capture_output=True,
            text=True,
        )
        print(process.stdout)
        count = process.stdout.split("\n")[0]
        self.assertEqual(count, "\x1b[94mCount: 1\x1b[0m")

    def test_J_kill_experiment_not_running(self):
        "Executing CMD kill not running experiment."
        print(self.shortDescription())
        process = subprocess.run(
            [
                "pharmacelera-cmd",
                "--option",
                "kill",
                "--id",
                self.__class__.exp_id,
                "--port",
                self.statics["PORT"],
                "--uri",
                self.statics["URI"],
            ],
            text=True,
            capture_output=True,
        )
        print(process.stdout)
        out_response = process.stdout.split("\n")[-3]
        self.assertEqual(out_response, "{'msg': 'No running/pending experiment found.'}")

    def test_K_download_experiment(self):
        "Executing CMD download finished experiment."
        print(self.shortDescription())
        process = subprocess.run(
            [
                "pharmacelera-cmd",
                "--option",
                "download",
                "--id",
                self.__class__.exp_id,
                "--port",
                self.statics["PORT"],
                "--uri",
                self.statics["URI"],
            ],
            text=True,
            capture_output=True,
        )
        print(process.stdout)
        generated_output = False
        for file in glob.glob(str(self.path) + "/*"):
            if file.find("test_case_cmd_") > -1:
                generated_output = True
                self.remove(file)
            if file.endswith("_TEST_MODEL") or file.endswith("tmp"):
                shutil.rmtree(file, ignore_errors=True)
        self.assertTrue(generated_output)

    def test_L_remove_experiment(self):
        "Execute CMD remove experiment."
        print(self.shortDescription())
        process = subprocess.run(
            [
                "pharmacelera-cmd",
                "--option",
                "remove",
                "--id",
                self.__class__.exp_id,
                "--port",
                self.statics["PORT"],
                "--uri",
                self.statics["URI"],
            ],
            text=True,
            capture_output=True,
        )
        print(process.stdout)
        process = subprocess.run(
            [
                "pharmacelera-cmd",
                "--id",
                self.__class__.exp_id,
                "--port",
                self.statics["PORT"],
                "--uri",
                self.statics["URI"],
            ],
            capture_output=True,
            text=True,
        )
        out_response = process.stdout.split("\n")[0]
        self.assertEqual(out_response, "\x1b[91mNo experiment found with the given id.\x1b[0m")

    def test_M_upload_no_parameter(self):
        """Execute upload binary command without parameter."""
        print(self.shortDescription())
        process = subprocess.run(["pharmacelera-cmd", "--option", "upload"], capture_output=True, text=True)
        out_response = process.stdout.split("\n")
        self.assertEqual(out_response[0], "\x1b[1mUploading binary and/or license ...\x1b[0m")
        self.assertEqual(out_response[1], "\x1b[91mPlease define binary and/or license path and filename.\x1b[0m")

    def test_N_upload_send_wrong_binary_type(self):
        """Execute upload binary command with wrong --type command."""
        print(self.shortDescription())
        process = subprocess.run(
            ["pharmacelera-cmd", "--option", "upload", "--type", "pharmqsa", "--binary", "notexists"],
            capture_output=True,
            text=True,
        )
        out_response = process.stdout.split("\n")[-2]
        self.assertEqual(out_response, "\x1b[91mBinary type is wrong, must be pharmqsar or pharmscreen")

    def test_O_upload_wrong_binary_path(self):
        """Execute upload binary command with a binary path that does not exists."""
        print(self.shortDescription())
        process = subprocess.run(
            ["pharmacelera-cmd", "--option", "upload", "--binary", "notexists"], capture_output=True, text=True
        )
        out_response = process.stdout.split("\n")[-2]
        self.assertEqual(
            out_response,
            "\x1b[91mCould not find the binary, please make sure binary is in the current folder or use absolute path.\x1b[0m",
        )

    def test_O_upload_binary_correctly(self):
        """Execute upload binary command with a binary path that does not exists."""
        print(self.shortDescription())
        process = subprocess.run(
            [
                "pharmacelera-cmd",
                "--option",
                "upload",
                "--binary",
                "testMocks/pharmqsar",
                "--port",
                self.statics["PORT"],
                "--uri",
                self.statics["URI"],
            ],
            capture_output=True,
            text=True,
        )
        out_response = process.stdout.split("\n")[1]
        self.assertEqual(
            out_response,
            "\x1b[92mUploaded.\x1b[0m",
        )

    def execute(self, cmd):
        """Executing command line with subprocess."""
        popen = subprocess.Popen(cmd, stdout=subprocess.PIPE, universal_newlines=True)
        for stdout_line in iter(popen.stdout.readline, ""):
            yield stdout_line
        popen.stdout.close()
        return_code = popen.wait()
        if return_code:
            raise subprocess.CalledProcessError(return_code, cmd)

    def remove(self, *args):
        """Remove files and ignore error.
        args: iterable strings
        """
        for file in args:
            try:
                os.remove(os.path.join(self.path, file))
            except OSError:
                pass



import os
import sys
import yaml
import boto3
import shutil
import zipfile
from typing import List
from pathlib import Path
from posixpath import isabs
from genericpath import isfile

#
# Copyright (c) 2021 Pharmacelera S.L.
# All rights reserved.
#
# Description: Support functions for experiment post-processing
#


class Colors:
    BOLD = "\033[1m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    OKBLUE = "\033[94m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    UNDERLINE = "\033[4m"


class Utils:
    # statics - - - - -
    out: str = "outputs"
    inputs: str = "inputs"
    result: str = "results.txt"
    log_file: str = "experiment.log"
    coef_file: str = "experiment.coef"

    def get_arg(self):
        """Parse binary name from the shell."""
        args = sys.argv
        if len(args) < 2:
            print("Please enter the binary name to upload: (pharmscreen or pharmqsar)")
            sys.exit()
        binary = args[1]
        if binary != "pharmscreen" and binary != "pharmqsar":
            print("Incorrect name, please use: (pharmscreen or pharmqsar)")
            sys.exit()
        return binary

    def get_name(self, command: str):
        """Extract experiment name from command line.
        Args:
            command (str): command line to execute the experiment
        """
        if command:
            command = command.split(" ")
            index = command.index("--name") + 1
            return command[index]
        return None

    def define_files(self, type: str, binary: str, license: str) -> List:
        """Define binary and license file names.
        Args:
            type (str): type of the binary either pharmscreen or pharmqsar
            binary (str): path and filename of the binary file
            license (str): path and filename of the license file

        Returns:
            files (list): list of the opened binary and license file streams.
        """
        files = []
        abs_path: str = Path().absolute()

        if type != "pharmqsar" and type != "pharmscreen":
            print(
                f"{Colors.FAIL}Binary type is wrong, must be pharmqsar or pharmscreen"
            )
            return files

        if binary:
            if not isabs(binary):
                binary = os.path.join(abs_path, binary)
            if not isfile(binary):
                print(
                    f"{Colors.FAIL}Could not find the binary, please make sure binary is in the current folder or use absolute path.{Colors.ENDC}"
                )
            else:
                files.append(
                    (
                        type,
                        (
                            type,
                            open(binary, "rb"),
                            "application/octet-stream",
                        ),
                    )
                )
        if license:
            if not isabs(license):
                license = os.path.join(abs_path, license)
            if not isfile(license):
                print(
                    f"{Colors.FAIL}Could not find the license, please make sure license is in the current folder or use absolute path.{Colors.ENDC}"
                )
            else:
                files.append(
                    (
                        f"{type}_license",
                        (
                            f"{type}_license",
                            open(license, "rb"),
                            "application/octet-stream",
                        ),
                    )
                )
        return files

    def best_model_selector(self, folder_path, statics, config=None):
        """Unzip all output files and list all log files,
        then parse line by line to extract the best component among all experiments
        """
        all_components = []
        self.best = statics["MODEL_NAME"]
        self.upload = statics["UPLOAD_S3"]
        if self.upload == True:
            self.s3_bucket = statics["S3_BUCKET"]
            self.s3_folder = statics["S3_FOLDER"]
        outputs = os.path.join(folder_path, self.out, self.best)
        results = os.path.join(folder_path, self.best)
        Path(results).mkdir(parents=True, exist_ok=True)
        result_file = open(os.path.join(results, self.result), "w")
        result_file.write(self.copyright())
        self.unzip_results(outputs)
        for root, _, files in os.walk(outputs):
            path = root.split(os.sep)
            vs_files = []
            for f in files:
                if f.endswith((".pdb", ".va0", ".coef")):
                    vs_files = files
            for file in vs_files:
                if file == self.log_file:
                    result_file.write(os.path.basename(root) + "\n")
                    filename = os.path.join("/".join(path), file)
                    best_q2 = 0.0
                    best_r2 = 0.0
                    num = 0
                    with open(filename, "r+") as f:
                        for line in f:
                            if line.startswith("test r2="):
                                r2 = float(line.split(" ")[2])
                                test_r2 = r2 if r2 > 0 else best_r2
                            if line.startswith("q2="):
                                q2 = float(line.split(" ")[1])
                                if q2 > best_q2 and (q2 - best_q2) > 0.01:
                                    best_q2 = q2
                                    num = next(f).split(" ")[-1].strip()
                                    best_r2 = test_r2
                    all_components.append(
                        {
                            "q2": best_q2,
                            "r2": best_r2,
                            "sum": best_q2 + best_r2,
                            "n": num,
                            "folder": path[-1],
                        }
                    )
                    result_file.write(
                        f"Best q2: {best_q2}\
                        \nBest r2: {best_r2}\
                        \nNumber of components: {num}\n"
                    )
        sorted_by_q2 = sorted(all_components, key=lambda k: k["q2"], reverse=True)
        sorted_by_r2 = sorted(all_components, key=lambda k: k["r2"], reverse=True)
        index_sum = sys.maxsize
        selected = {}
        for q2_index, content in enumerate(sorted_by_q2):
            r2_index = next(
                (i for (i, d) in enumerate(sorted_by_r2) if d["r2"] == content["r2"]),
                None,
            )
            qr_sum = q2_index + r2_index
            if qr_sum < index_sum:
                index_sum = qr_sum
                selected = sorted_by_q2[q2_index]
            elif qr_sum == index_sum:
                q2_index = (
                    q2_index - 1
                    if selected["sum"] > sorted_by_q2[q2_index]["sum"]
                    else q2_index
                )
                selected = sorted_by_q2[q2_index]
        result_file.write(self.headline())
        result_file.write(self.get_row(selected))
        result_file.close()
        if self.upload == True and selected:
            folder = selected["folder"]
            self.create_next_yaml(selected, config)
            self.prepare_files_and_upload(folder_path, folder)
        self.remove_uncompressed()

    def create_next_yaml(self, selected, config):
        """Manual customization of new yaml config for next experiments."""
        chosen_exp = [e for e in config.values() if e["name"] == selected["folder"]].pop()
        prep_exp = [
            e for e in config.values()
            if e["name"] == chosen_exp["files"]["input"][4:-4]
        ]
        exp = dict()
        if prep_exp:
            prep_exp = prep_exp.pop()
            prep_exp["mode"] = "predict"
            prep_exp["repcomp"] = selected["n"]
            prep_exp["files"]["repcoef"] = "inputs/experiment.coef"
            prep_exp["files"]["repgrid"] = "inputs/grid_out.pdb"
            if "ref" in prep_exp["files"]:
                prep_exp["files"]["ref"] = prep_exp["files"]["ref"].replace(
                    "tmp/", "inputs/tmp/"
                )
            exp = {"experiment1": prep_exp}
        with open("custom.yaml", "w") as file:
            yaml.dump(exp, file)

    def unzip_results(self, outputs):
        folder, _, filenames = next(os.walk(outputs))
        for f in filenames:
            if f.endswith(".zip"):
                with zipfile.ZipFile(os.path.join(folder, f)) as zipped:
                    path = os.path.join(outputs, f.split(".")[0])
                    Path(path).mkdir(parents=True, exist_ok=True)
                    zipped.extractall(path)

    def prepare_files_and_upload(self, path, folder):
        for root, _, files in os.walk(os.path.join(path, self.out, self.best, folder)):
            for file in files:
                if file.endswith((".coef", ".pdb", ".log")):
                    dest = file
                    if file.endswith(".coef"):
                        dest = self.coef_file
                    shutil.copyfile(
                        os.path.join(root, file), os.path.join(path, self.best, dest)
                    )
        shutil.move(
            os.path.join(path, "custom.yaml"),
            os.path.join(path, self.best, "custom.yaml"),
        )
        try:
            shutil.move(os.path.join(path, "tmp"), os.path.join(path, self.best, "tmp"))
        except Exception:
            pass
        shutil.make_archive(self.best, "zip", os.path.join(path, self.best))
        shutil.rmtree(os.path.join(path, self.best), ignore_errors=True)
        self.upload_s3(path, self.best)

    def upload_s3(self, path, file):
        if self.upload:
            file = file + ".zip"
            s3 = boto3.resource("s3")
            s3.Bucket(self.s3_bucket).upload_file(file, self.s3_folder + "/" + file)
            print("Uploaded to S3")

    def get_row(self, selected):
        return f"Experiment folder: {selected.get('folder'), None}\
                \nBest q2: {selected.get('q2'), None}\
                \nBest test r2: {selected.get('r2', None)}\
                \nNumber of components: {selected.get('n')} \n\n"

    def remove_uncompressed(self):
        outputs = os.path.join(self.out, self.best)
        for path in Path(outputs).glob("**/*"):
            if path.is_dir():
                shutil.rmtree(path)

    @staticmethod
    def headline():
        return "\n" + "-" * 30 + "Selected Configuration" + "-" * 30 + "\n"

    @staticmethod
    def copyright():
        return (
            "PHARMQSAR MODEL SELECTION\
            \nPHARMACELERA S.L.\
            \n"
            + "-" * 80
            + "\n"
        )



import sys
import logging
import botocore
import requests
from .utils import Colors


class Errors:
    def __init__(self) -> None:
        pass

    def requestException(self, error):
        if isinstance(error, requests.exceptions.RequestException):
            logging.error("Request failed")
            logging.error(f"{Colors.FAIL}{str(error)}")

    def S3Exception(self, error):
        if isinstance(error, botocore.exceptions.ClientError):
            logging.error(f"{Colors.FAIL}{str(error)}{Colors.ENDC}")
            if "Not Found" in str(error):
                logging.info(f"{Colors.FAIL}File from S3 not found.{Colors.ENDC}")
        elif isinstance(error, botocore.exceptions.NoCredentialsError):
            logging.error(f"{Colors.FAIL}AWS credentials not found.{Colors.ENDC}")
        else:
            logging.error(f"{Colors.FAIL}{Colors.BOLD}Unhandled error occured: {type(error).__name__}{Colors.ENDC}")

        logging.error(f"{Colors.FAIL}Download from S3 failed.{Colors.ENDC}")

    def fileNameException(self):
        logging.error(
            f"{Colors.FAIL}If 'GET_CONF_FROM_S3' parameter is defined as true, then 'FILENAME' must be defined in config yaml.{Colors.ENDC}"
        )
        sys.exit(0)

    def bucketNameException(self):
        logging.error(f"{Colors.FAIL}S3_BUCKET parameter is not defined in configuration yaml.{Colors.ENDC}")
        sys.exit(0)

    def S3FolderPathException(self):
        logging.error(f"{Colors.FAIL}S3_FOLDER parameter is not defined in configuration yaml.{Colors.ENDC}")
        sys.exit(0)

    def inputFileNameException(self):
        logging.error(
            f"{Colors.FAIL}If 'GET_CONF_FROM_S3' parameter is defined as true, then 'EXP_1_INPUT' must be defined in config yaml.{Colors.ENDC}"
        )
        sys.exit(0)

    def inputFilesException(self, fileConf):
        logging.error(
            f"{Colors.FAIL}Input files could not be loaded. Please check the path and name of the files.{Colors.ENDC}\n{str(fileConf)}"
        )
        sys.exit(0)

    def downloadFilesException(self, exp):
        logging.error(f"{Colors.FAIL}Download files failed.{Colors.ENDC}\n{str(exp.message)}")
        sys.exit(0)

    def yamlParseException(self, exp):
        logging.error(f"{Colors.FAIL}Yaml config file is malformed.{Colors.ENDC}")


from setuptools import setup, find_packages

setup(
    name="pharmacelera_utils",
    version="0.1.53",
    description="Utility functions to execute pharmacelera scripts",
    url="https://bitbucket.org/pharmacelera/pharmqsar-examples.git",
    author="Pharmacelera developers",
    author_email="support@pharmacelera.com",
    license="Mozilla Public License Version 2.0",
    packages=find_packages(where="src"),
    install_requires=["requests==2.31.0", "PyYAML==6.0.1", "boto3==1.28.62", "botocore==1.31.62"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    py_modules=[
        "launch",
        "services",
        "utils",
        "errors",
        "commands",
        "__init__",
    ],
    package_dir={"": "src"},
    entry_points={
        "console_scripts": [
            "pharmacelera-launch=pharmacelera_utils.launch:run",
            "pharmacelera-cmd=pharmacelera_utils.commands:run",
        ]
    },
)